﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Anwendung                                                                           //
//  _____________________________________________________                               //
//                                                                                      //
//  Copyright (c) 2019 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include <math.h>
#include "Cy4Net.h"
#include "Cy4TypePlate.h"
#include "Conf.h"
#include "Application.h"
#include "Board.h"
#include "Storing.h"
#include "Cy4Link.h"
#include "Timer.h"
#include "DateTime.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static  Timer  serverQueryInterval[2] = { TIMER_SECONDS(0), TIMER_SECONDS(0) };
static  Timer  extraOnTimer[2]        = { TIMER_SECONDS(0), TIMER_SECONDS(0) };
static  SInt08 ambientBrightness      = CFG_INVALID_BRIGHTNESS;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static PowerSwitch
DestinationOf(const Cy4Message* message)
{
    if (CY4MESSAGE_Location(message) < 0x100200) return POWER_SWITCH_1;
    else                                         return POWER_SWITCH_2;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ValidateAmbientBrightness(void)
{
    Cy4Link  serverLink;
    STORING_Load(STO_ServerLink[1], &serverLink);
    bool enabled;
    STORING_Load(STO_AcceptBroadcast[1], &enabled);
    if (!enabled && !CY4LINK_IsValid(&serverLink))
    {   // Keine gültiger Umgebungshelligkeitsserver und Rundrufe werden ignoriert
        ambientBrightness = CFG_INVALID_BRIGHTNESS;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
void
StartExtraOn(PowerSwitch powerSwitch)
{
    UInt08 i = (UInt08)powerSwitch;
    UInt08 duration;
    STORING_Load(STO_ExtraOnDuration[i], &duration);
    TIMER_Start(&extraOnTimer[i], duration*60);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
StopExtraOn(PowerSwitch powerSwitch)
{
    UInt08 i = (UInt08)powerSwitch;
    TIMER_Stop(&extraOnTimer[i]);
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
IsExtraOn(PowerSwitch powerSwitch)
{
    UInt08 i = (UInt08)powerSwitch;
    return !TIMER_HasExpired(&extraOnTimer[i]);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Nachrichtenausstatter  ///////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupQueryTimeMessage(Cy4Message* message)
{
    Cy4Link serverLink;
    STORING_Load(STO_ServerLink[0], &serverLink);
    if (CY4LINK_IsValid(&serverLink))
    {   CY4MESSAGE_SetupRead(message, serverLink.address, serverLink.location, 7);
    }
    else
    {   CY4MESSAGE_Invalidate(message);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupQueryBrightnessMessage(Cy4Message* message)
{
    Cy4Link serverLink;
    STORING_Load(STO_ServerLink[1], &serverLink);
    if (CY4LINK_IsValid(&serverLink))
    {   CY4MESSAGE_SetupRead(message, serverLink.address, serverLink.location, 1);
    }
    else
    {   CY4MESSAGE_Invalidate(message);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Antwort-Behandler  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
ResponseOfQueryTime(const Cy4Message* message)
{
    if (CY4MESSAGE_IsValid(message))
    {   DateTime dateTime;
        DATETIME_FromData(CY4MESSAGE_ConstData(message), &dateTime);
        if (DATETIME_IsValid(&dateTime)) BOARD_SetClock(&dateTime);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ResponseOfQueryBrightness(const Cy4Message* message)
{
    if (CY4MESSAGE_IsValid(message))
    {   ambientBrightness = (SInt08)CY4MESSAGE_ConstData(message)[0];
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Anfrage-Bearbeiter  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: PowerSwitchMode (read write)                                                //
//                                                                                      //
//  Beschreibung: Betriebsart des Leistungsschalters.                                   //
//                PowerSwitchMode ist ein Gemeinschaftsbearbeiter der Leistungsschalter //
//                1 und 2.                                                              //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |PSM|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//    [PSM] Betriebsart: 0 = Aus (Voreinstellung),                                      //
//                       1 = Ein,                                                       //
//                       2 = Zeitautomatik,                                             //
//                       3 = Helligkeitsautomatik,                                      //
//                       x = Wird ignoriert                                             //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnPowerSwitchMode(const Cy4Service* service, Cy4Message* message)
{
    UInt08 i = (UInt08)DestinationOf(message);
    PowerSwitchMode mode;
    STORING_Load(STO_PowerSwitchMode[i], &mode);
    UInt08 data = (UInt08)mode;
    CY4SERVICE_TransferPayload(service, message, &data);
    if (CY4MESSAGE_Access(message) == CY4MESSAGE_WRITE)
    {   if (data < 4)
        {   mode = (PowerSwitchMode)data;
            STORING_Save(&mode, STO_PowerSwitchMode[i]);
            PowerSwitch powerSwitch = (PowerSwitch)i;
            StopExtraOn(powerSwitch);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: PowerSwitchState (read only)                                                //
//                                                                                      //
//  Beschreibung: Liefert den Zustand des Leistungsschalters.                           //
//                PowerSwitchState ist ein Gemeinschaftsbearbeiter der Leistungs-       //
//                schalter 1 und 2.                                                     //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |PSS|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//    [PSS] Leistungsschalter: 0 = Aus, 1 = Ein.                                        //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnPowerSwitchState(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    PowerSwitch powerSwitch = DestinationOf(message);
    CY4MESSAGE_Data(message)[0] = BOARD_PowerSwitch(powerSwitch);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ExtraOn (write only)                                                        //
//                                                                                      //
//  Beschreibung: Auslöser, um unabhängig von der aktuellen Betriebsart den Leistungs-  //
//                schalter für eine bestimmte Dauer (s. ExtraOnDuration.DUR) einzu-     //
//                schalten.                                                             //
//                ExtraOn ist ein Gemeinschaftsbearbeiter der Leistungsschalter 1 und   //
//                2.                                                                    //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EXO|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//                                                                                      //
//  [EXO] Auslöser zur betriebsartunabhängigen zeitbegrenzten Einschaltung des          //
//        Leistungsschalter: 5Eh ('^').                                                 //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnExtraOn(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    PowerSwitch powerSwitch = DestinationOf(message);
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) StartExtraOn(powerSwitch);
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ExtraOnDuration (read write only)                                           //
//                                                                                      //
//  Beschreibung: Dauer der Extraeinschaltung. Siehe ExtraOn.EXO.                       //
//                ExtraOnDuration ist ein Gemeinschaftsbearbeiter der Leistungsschalter //
//                1 und 2.                                                              //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |DUR|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//                                                                                      //
//  [DUR] Dauer der betriebsartunabhängigen Einschaltung des Leistungsschalter: 1 ..    //
//        240 [min].                                                                    //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnExtraOnDuration(const Cy4Service* service, Cy4Message* message)
{
    UInt08 i = (UInt08)DestinationOf(message);
    UInt08 data;
    STORING_Load(STO_ExtraOnDuration[i], &data);
    CY4SERVICE_TransferPayload(service, message, &data);
    if (CY4MESSAGE_Access(message) == CY4MESSAGE_WRITE)
    {   if (data <   1) data =   1;
        if (data > 240) data = 240;
        STORING_Save(&data, STO_ExtraOnDuration[i]);
        PowerSwitch powerSwitch = (PowerSwitch)i;
        if (IsExtraOn(powerSwitch)) StartExtraOn(powerSwitch);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: TimerModeParameter (read write)                                             //
//                                                                                      //
//  Beschreibung: Zugriff auf die Parameter der Zeitautomatik. Überschneiden sich       //
//                Zeiten der Zeitgeber, dominiert der eingeschaltete Zustand.           //
//                TimerModeParameter ist ein Gemeinschaftsbearbeiter für alle 3 Zeit-   //
//                geber der Leistungsschalter 1 und 2.                                  //
//                                                                                      //
//              0   1   2   3   4                                                       //
//            +---+---+---+---+---+  5   6   7   8   9   10  11                         //
//  Belegung: |ACT|STH|STM|ETH|ETM|+---+---+---+---+---+---+---+                        //
//            +---+---+---+---+---+|MON|TUE|WED|THU|FRI|SAT|SUN|                        //
//                                 +---+---+---+---+---+---+---+                        //
//                                                                                      //
//    [ACT] Zeitgeber Aktiv: 0 = Aus, 1 = Ein, x = Wird ignoriert.                      //
//    [STH] Einschaltzeitpunkt Stunde: 0 .. 23 [h], oder 255.                           //
//    [STM] Einschaltzeitpunkt Minute: 0 .. 59 [m], oder 255.                           //
//    [ETH] Abschaltzeitpunkt Stunde:  0 .. 23 [h], oder 255.                           //
//    [ETM] Abschaltzeitpunkt Minute:  0 .. 59 [m], oder 255.                           //
//    [MON] Freigabe für Montag:     0 = Aus 1 = Ein, x = Wird ignoriert.               //
//    [TUE] Freigabe für Dienstag:   0 = Aus 1 = Ein, x = Wird ignoriert.               //
//    [WED] Freigabe für Mittwoch:   0 = Aus 1 = Ein, x = Wird ignoriert.               //
//    [THU] Freigabe für Donnerstag: 0 = Aus 1 = Ein, x = Wird ignoriert.               //
//    [FRI] Freigabe für Freitag:    0 = Aus 1 = Ein, x = Wird ignoriert.               //
//    [SAT] Freigabe für Samstag:    0 = Aus 1 = Ein, x = Wird ignoriert.               //
//    [SUN] Freigabe für Sonntag:    0 = Aus 1 = Ein, x = Wird ignoriert.               //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnTimerModeParameter(const Cy4Service* service, Cy4Message* message)
{
    UInt08 i = (UInt08)DestinationOf(message);                    // i. Leistungsschalter
    UInt08 n = ((CY4MESSAGE_Location(message) & 0x0000F0)>>4) -2; // n. Zeitgeber
    AutoTimer autoTimer;
    STORING_Load(STO_AutoTimer[i][n], &autoTimer);
    UInt08 data[12];
    data[0] = (UInt08)autoTimer.enabled;
    data[1] = autoTimer.startTime.hour;
    data[2] = autoTimer.startTime.minute;
    data[3] = autoTimer.endTime.hour;
    data[4] = autoTimer.endTime.minute;
    for (UInt08 d=0; d<7; d++) data[5+d] = (UInt08)autoTimer.weekDay[d];
    CY4SERVICE_TransferPayload(service, message, data);
    if (data[0] <  2) autoTimer.enabled = (bool)data[0];
    if (data[1] < 24) autoTimer.startTime.hour   = data[1];
    if (data[2] < 60) autoTimer.startTime.minute = data[2];
    autoTimer.startTime.second = 0;
    if (data[3] < 24) autoTimer.endTime.hour   = data[3];
    if (data[4] < 60) autoTimer.endTime.minute = data[4];
    autoTimer.endTime.second = 0;
    for (UInt08 d=0; d<7; d++)
    {   if (data[5+d] < 2) autoTimer.weekDay[d] = (bool)data[5+d];
    }
    STORING_Save(&autoTimer, STO_AutoTimer[i][n]);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: BrightnessModeParameter (read write)                                        //
//                                                                                      //
//  Beschreibung: Zugriff auf die Parameter der Helligkeitsautomatik.                   //
//                BrightnessModeParameter ist ein Gemeinschaftsbearbeiter der           //
//                Leistungsschalter 1 und 2.                                            //
//                                                                                      //
//              0   1   2   3                                                           //
//            +---+---+---+---+                                                         //
//  Belegung: |ACT|THB|THD|ONT|                                                         //
//            +---+---+---+---+                                                         //
//                                                                                      //
//    [ACT] Aktivität                                                                   //
//          0 = Ein wenn dunkel, aus wenn hell (Voreinstellung).                        //
//          1 = Ein wenn hell, aus wenn dunkel.                                         //
//          x = Ohne Wirkung                                                            //
//    [THB] Helligkeits-Schwellenwert: 2 .. 99 %. <THB> muss über <THD> liegen.         //
//    [THD] Dunkelheits-Schwellenwert: 1 .. 98 %. <THD> muss unter <THB> liegen.        //
//    [ONT] Automatische Abschaltung nach (in 10-Minuten Schritten): 0 .. 96 [10m],     //
//          0 = Keine Abschaltung .                                                     //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnBrightnessModeParameter(const Cy4Service* service, Cy4Message* message)
{
    UInt08 data[5];
    UInt08 i = (UInt08)DestinationOf(message);
    BrightnessMode mode;
    STORING_Load(STO_BrightnessMode[i], &mode);
    data[0] = (UInt08)mode;
    STORING_Load(STO_LightnessThreshold[i], &data[1]);
    STORING_Load(STO_DarknessThreshold[i], &data[2]);
    STORING_Load(STO_BrightnessAutoOffTime[i], &data[3]);
    CY4SERVICE_TransferPayload(service, message, data);
    if (data[0] < 2)
    {   mode = (BrightnessMode)data[0];
        STORING_Save(&mode, STO_BrightnessMode[i]);
    }
    if (data[2] > data[1]) data[2] = data[1]-1;
    if (data[1] <  2) data[1] =  2;
    if (data[1] > 99) data[1] = 99;
    STORING_Save(&data[1], STO_LightnessThreshold[i]);
    if (data[2] <  1) data[2] =  1;
    if (data[2] > 98) data[2] = 98;
    STORING_Save(&data[2], STO_DarknessThreshold[i]);
    if (data[4] > 96) data[4] = 96;
    STORING_Save(&data[3], STO_BrightnessAutoOffTime[i]);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: DateTime (read write)                                                       //
//                                                                                      //
//  Beschreibung: Ermöglicht den Zugriff auf die Systemzeit des Gerätes. Der Behandler  //
//                wird für den lokalen Dienst wie auch für den Rundruf eingesetzt.      //
//                                                                                      //
//              0   1   2   3   4   5   6                                               //
//            +---+---+---+---+---+---+---+                                             //
//  Belegung: |YR1|YR2|MON|DAY|HUR|MIN|SEC|                                             //
//            +---+---+---+---+---+---+---+                                             //
//                                                                                      //
//    [YRx] Jahr:    1700 .. 9000 , (0xFFFF)                                            //
//    [MON] Monat:      1 ..   12 , (0xFF)                                              //
//    [DAY] Tag:        1 ..   31 , (0xFF)                                              //
//    [HUR] Stunde:     0 ..   23 , (0xFF)                                              //
//    [MIN] Minute:     0 ..   59 , (0xFF)                                              //
//    [SEC] Sekunde:    0 ..   59 , (0xFF)                                              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnDateTime(const Cy4Service* service, Cy4Message* message)
{
    if (CY4MESSAGE_Location(message) == 0xE00020)
    {   bool broadcastAccepted;
        STORING_Load(STO_AcceptBroadcast[0], &broadcastAccepted);
        if (!broadcastAccepted) return;
    }
    UInt08 data[7];
    {   DateTime  dateTime;
        BOARD_Clock(&dateTime);
        DATETIME_ToData(&dateTime, data);
    }
    CY4SERVICE_TransferPayload(service, message, data);
    if (CY4MESSAGE_Access(message) == CY4MESSAGE_WRITE)
    {   DateTime dateTime;
        DATETIME_FromData(data, &dateTime);
        BOARD_SetClock(&dateTime);
        TIMER_Start(&serverQueryInterval[0], CFG_TIME_QUERY_INTERVAL*60);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: AmbientBrightness (read write)                                              //
//                                                                                      //
//  Beschreibung: Ermöglicht Zugriff auf die vom System genutzte Umgebungshelligkeit.   //
//                Der Behandler wird für den lokalen Dienst wie auch für den Rundruf    //
//                eingesetzt.                                                           //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EBS|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EBS] Umgebungshelligkeit: 0 .. 100 % oder -128.                                    //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnAmbientBrightness(const Cy4Service* service, Cy4Message* message)
{
    if (CY4MESSAGE_Location(message) == 0xE00060)
    {   bool broadcastAccepted;
        STORING_Load(STO_AcceptBroadcast[1], &broadcastAccepted);
        if (!broadcastAccepted) return;
    }
    UInt08 data = ambientBrightness;
    CY4SERVICE_TransferPayload(service, message, &data);
    ambientBrightness = data;
    if (CY4MESSAGE_Access(message) == CY4MESSAGE_WRITE)
    {   TIMER_Start(&serverQueryInterval[1], CFG_BRIGHTNESS_QUERY_INTERVAL*60);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: AcceptBroardcasts (read write)                                              //
//                                                                                      //
//  Beschreibung: Mit diesem Dienst kann festgelegt werden, ob das Gerät Rundmitteil-   //
//                ungen an E00020h (Zeit) und E00060h (Umgebungshelligkeit) entgegen-   //
//                nimmt oder nicht.                                                     //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |ATB|ABB|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [ATB] Zeit-Rundmitteilung: 1 = akzeptieren, 0 = ignorieren.                         //
//  [ABB] Umgebungshelligkeit-Rundmitteilung: 1 = akzeptieren, 0 = ignorieren.          //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnAcceptBroardcasts(const Cy4Service* service, Cy4Message* message)
{
    UInt08 data[2];
    for (UInt08 i=0; i<2; i++)
    {   bool enable;
        STORING_Load(STO_AcceptBroadcast[i], &enable);
        data[i] = (UInt08)enable;
    }
    CY4SERVICE_TransferPayload(service, message, data);
    for (UInt08 i=0; i<2; i++)
    {   bool enable = (bool)data[i];
        STORING_Save(&enable, STO_AcceptBroadcast[i]);
    }
    if (CY4MESSAGE_Access(message) == CY4MESSAGE_WRITE)
    {   ValidateAmbientBrightness();
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ServerLink (read write)                                                     //
//                                                                                      //
//  Beschreibung: Adresse und Dienstort der Server für Zeit und Umgebungshelligkeit.    //
//                                                                                      //
//              0   1   2   3   4   5   6   7   8   9                                   //
//            +---+---+---+---+---+---+---+---+---+---+                                 //
//  Belegung: |TSN|TSD|LC1|LC2|LC3|BSN|BSD|EC1|EC2|EC3|                                 //
//            +---+---+---+---+---+---+---+---+---+---+                                 //
//                                                                                      //
//  [TSN] Zeitserver: Cy4-Netzadresse: 1 .. 255.                                        //
//  [TSD] Zeitserver: Cy4-Geräteadresse: 1 .. 255.                                      //
//  [LCx] Zeitserver: Dienstort der abzufragenden Zeitinformation: 100000h .. FFFF00h,  //
//        FFFFFFh wenn nicht gesetzt.                                                   //
//  [BSN] Umgebungshelligkeitsserver: Cy4-Netzadresse: 1 .. 255.                        //
//  [BSD] Umgebungshelligkeitsserver: Cy4-Geräteadresse: 1 .. 255.                      //
//  [ECx] Umgebungshelligkeitsserver: Dienstort der abzufragenden helligkeit: 100000h   //
//        .. FFFF00h, FFFFFFh wenn nicht gesetzt .                                      //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnServerLink(const Cy4Service* service, Cy4Message* message)
{
    UInt08  data[10];
    for (UInt08 i=0; i<2; i++)
    {   Cy4Link  serverLink;
        STORING_Load(STO_ServerLink[i], &serverLink);
        CY4LINK_ToData(&serverLink, &data[5*i]);
    }
    CY4SERVICE_TransferPayload(service, message, data);
    for (UInt08 i=0; i<2; i++)
    {   Cy4Link  serverLink;
        CY4LINK_FromData(&data[5*i], &serverLink);
        STORING_Save(&serverLink, STO_ServerLink[i]);
        if (CY4MESSAGE_Access(message) == CY4MESSAGE_WRITE)
        {   if (CY4LINK_IsValid(&serverLink)) TIMER_Start(&serverQueryInterval[i], 3);
            ValidateAmbientBrightness();
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: InfoLedSignal (read write)                                                  //
//                                                                                      //
//  Beschreibung: Legt die Aufgabe der grünen Info-LEDs fest. Folgende Lichtsignale     //
//                können eingestellt werden:                                            //
//                                                                                      //
//                       0 = Kein Signal, Ausgeschaltet.                                //
//                       1 = Leuchtet durchgehend.                                      //
//                       2 = Herzschlag-Signal.                                         //
//                       3 = Zeigt Aktivität auf dem Bus.                               //
//                       x = Wirkungslos.                                               //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |SIG|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [SIG] Lichtsignale der grünen Info-LED, Voreingestellt ist 3.                       //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnInfoLedSignal(const Cy4Service* service, Cy4Message* message)
{
    UInt08 data[1];
    {   InfoLedSignal signal;
        STORING_Load(STO_InfoLedSignal, &signal);
        data[0] = (UInt08)signal;
    }
    CY4SERVICE_TransferPayload(service, message, data);
    if (data[0] < 4)
    {   InfoLedSignal signal = (InfoLedSignal)data[0];
        STORING_Save(&signal, STO_InfoLedSignal);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: DeviceLabel (read write)                                                    //
//                                                                                      //
//  Beschreibung: Benutzerdefinierte Gerätebezeichnung.                                 //
//                                                                                      //
//              0   1   2   ...  59  60                                                 //
//            +---+---+---+-   -+---+---+                                               //
//  Belegung: |N01|N02|N03| ... |N60| 0 |                                               //
//            +---+---+---+-   -+---+---+                                               //
//                                                                                      //
//  [Nxx] Zeichen der allgemeinen Gerätebezeichnung: 0 .. 255 (ASCII/UTF-8)             //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnDeviceLabel(const Cy4Service* service, Cy4Message* message)
{
    UInt08  label[61];

    STORING_Load(STO_DeviceLabel, label);
    CY4SERVICE_TransferPayload(service, message, label);
    label[sizeof(label)-1] = '\0';
    STORING_Save(label, STO_DeviceLabel);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: FactorySettings (write only)                                                //
//                                                                                      //
//  Beschreibung: Zurückstellen aller Einstellungen auf den Ursprungszustand.           //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |FST|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [FST] Auslöser zur Wiederherstellung der Werksvoreinstellungen: 5Eh ('^').          //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnFactorySettings(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) STORING_RestoreDefaults();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: CallTypePlate (read only)                                            //
//                                                                                      //
//  Beschreibung: Geräte-Typenschild erfragen.                                          //
//                                                                                      //
//              0   1   2   3   4   5   6                                               //
//            +---+---+---+---+---+---+---+                                             //
//  Belegung: |AT2|AT1|AT0|AVE|BT2|BT1|BT0|                                             //
//            +---+---+---+---+---+---+---+                                             //
//                                                                                      //
//  [ATx] Anwendungstypennummer: Axxxxxh, xxxxx in BCD: 00001 .. 99999                  //
//  [AVE] Anwendungsversion:     01 .. 99, Haupt/Unternummer in BCD (Bsp: 12h = V1.2)   //
//  [BTx] Boardtypennummer:      Bxxxxxh, xxxxx in BCD: 00001 .. 99999                  //
//                                                                                      //
//                                                                                      //
//      Beispiel:                                                                       //
//                                                                                      //
//          +---+---+---+---+---+---+---+                                               //
//          |A0h|01h|23h|45h|B0h|67h|89h|      Anwendungstypennummer: 123               //
//          +---+---+---+---+---+---+---+      Anwendungsversion:     4.5               //
//                                             Boardtypennummer:      6789              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnCallTypePlate(const Cy4Service* service, Cy4Message* message)
{
    static const UInt08  typePlate[][7] PROGMEM = {  { 'C','y','4','C','o','d','e' },
                                                      CY4TYPEPLATE(DEV_APPL_TYPE,
                                                      DEV_APPL_VERS,
                                                      DEV_BOARD_TYPE)
                                                  };

    UInt08 j = (UInt08)(CY4MESSAGE_Location(message) - CY4SERVICE_Base(service));
    for (UInt08 i=0; i<CY4MESSAGE_Count(message); i++)
    {   CY4MESSAGE_Data(message)[i] = pgm_read_byte(&typePlate[1][j]);
        j++;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: EnterApplication (write only)                                        //
//                                                                                      //
//  Beschreibung: Startet die Anwendung neu.                                            //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EAP|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EAP] Auslöser zum (Neu-)Start der Applikation (Reset): 5Eh ('^').                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnEnterApplication(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) BOARD_Reset();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: EnterProgramLoader (write only)                                      //
//                                                                                      //
//  Beschreibung: Startet das System neu und kehrt im Programmlader-Modul zurück.       //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EPL|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EPL] Auslöser zum Eintritt in den Programmlader: 5Eh ('^').                        //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnEnterProgramLoader(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) BOARD_EnterPgmLdr();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: ChangeCy4Address (write only)                                        //
//                                                                                      //
//  Beschreibung: Wechsel der Geräteadresse.                                            //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |NET|DEV|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [NET] Neue Netzadresse:   1 .. 255                                                  //
//  [DEV] Neue Geräteadresse: 1 .. 255                                                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnChangeCy4Address(const Cy4Service* service, Cy4Message* message)
{
    UInt08  data[2];
    {   Cy4Address address = CY4NET_Address();
        data[0] = CY4ADDRESS_Net(&address);
        data[1] = CY4ADDRESS_Dev(&address);
    }
    CY4SERVICE_TransferPayload(service, message, data);
    {   Cy4Address address = CY4ADDRESS(data[0], data[1]);
        if (CY4ADDRESS_IsBroadcast(&address) || !CY4ADDRESS_IsValid(&address))
        {   CY4MESSAGE_Invalidate(message);
        }
        else
        {   CY4NET_SetAddress(address);
            STORING_SaveCy4SelfAdr(&address);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Verfahrensfunktionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProccedServerRequests(void)
{
    if (TIMER_HasExpired(&serverQueryInterval[0]))
    {   TIMER_Start(&serverQueryInterval[0], CFG_TIME_QUERY_INTERVAL*60);
        CY4NET_Post(SetupQueryTimeMessage, ResponseOfQueryTime);
    }
    if (TIMER_HasExpired(&serverQueryInterval[1]))
    {   TIMER_Start(&serverQueryInterval[1], CFG_BRIGHTNESS_QUERY_INTERVAL*60);
        CY4NET_Post(SetupQueryBrightnessMessage, ResponseOfQueryBrightness);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedLedSignaling(void)
{
    // Info-LED
    InfoLedSignal signal;
    STORING_Load(STO_InfoLedSignal, &signal);
    UInt16 ticks = BOARD_MSecTicks();
    bool ledOn;
    switch (signal)
    {   case INFOLEDSIGNAL_OFF:       ledOn = false;                  break;
        case INFOLEDSIGNAL_ON:        ledOn = true;                   break;
        case INFOLEDSIGNAL_HEARTBEAT: ledOn = (bool)((ticks & 0x380)==0x80 || (ticks & 0x380)==0x180); break;
        case INFOLEDSIGNAL_BUS_ACT:   ledOn = BOARD_BusActivity();    break;
        default:                      ledOn = false;
    }
    BOARD_SetLed(LED_INFO, ledOn);
    // Status-LEDs der Leistungsschalter
    static UInt08 state[2] = { 0 , 0 };
    static Timer  timer[2] = { TIMER_MILLISECONDS(100), TIMER_MILLISECONDS(100) };
    const  UInt08 pulses   = 4;
    for (UInt08 i=0; i<2; i++)
    {   if (state[i] == 0)
        {   if (BOARD_PowerSwitch((PowerSwitch)i)) state[i]++;
        }
        else if (state[i] == pulses*2+2)
        {   if (!BOARD_PowerSwitch((PowerSwitch)i)) state[i]++;
        }
        else if (state[i] == (pulses*2+2)*2)
        {   state[i] = 0;
        }
        else if (TIMER_RestartIfExpired(&timer[i]))
        {   BOARD_SetLed((Led)i, !BOARD_Led((Led)i));
            state[i]++;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedTimerAutomatic(PowerSwitch powerSwitch)
{
    DateTime dateTime;
    BOARD_Clock(&dateTime);
    if (!DATETIME_IsValid(&dateTime))
    {   BOARD_SetPowerSwitch(powerSwitch, false);
        return;
    }
    Time currentTime = dateTime.time;
    bool on = false;
    for (UInt08 n=0; n<3; n++)
    {   AutoTimer autoTimer;
        UInt08 i = (UInt08)powerSwitch;
        STORING_Load(STO_AutoTimer[i][n], &autoTimer);
        if (!autoTimer.enabled ||
            !TIME_IsValid(&autoTimer.startTime) ||
            !TIME_IsValid(&autoTimer.endTime))
        {   continue;
        }
        if (TIME_Compare(&autoTimer.startTime, &autoTimer.endTime) > 0)
        {   // Bsp: Von 23:10 bis 07:15
            on = TIME_Compare(&currentTime, &autoTimer.startTime) >= 0 ||
                 TIME_Compare(&currentTime, &autoTimer.endTime) < 0;

        }
        else
        {   // Bsp: Von 10:30 bis 15:10
            on = TIME_Compare(&currentTime, &autoTimer.startTime) >= 0 &&
                 TIME_Compare(&currentTime, &autoTimer.endTime) < 0;
        }
        if (on)
        {   DTWeekday weekday = DATE_WeekDay(&dateTime.date);
            if (weekday != DT_NOD)
            {   on = (bool)autoTimer.weekDay[(UInt08)weekday-1];
            }
        }
        if (on) break;  // "Ein" dominiert "Aus"
    }
    BOARD_SetPowerSwitch(powerSwitch, on);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedBrightnessAutomatic(PowerSwitch powerSwitch)
{
    UInt08 i = (UInt08)powerSwitch;
    if (ambientBrightness == CFG_INVALID_BRIGHTNESS)
    {   BOARD_SetPowerSwitch(powerSwitch, false);
        return;
    }
    // (1) Aktuellen Helligkeitszustand ermitteln
    enum
    {   LIGHT_STAGE,    // Zustand Hell
        INTERSTAGE,     // Zwischenzustand
        DARK_STAGE      // Zustand Dunkel
    } stage = INTERSTAGE;
    UInt08 lightnessThreshold;
    STORING_Load(STO_LightnessThreshold[i], &lightnessThreshold);
    UInt08 darknessThreshold;
    STORING_Load(STO_DarknessThreshold[i], &darknessThreshold);
    if      (ambientBrightness > lightnessThreshold) stage = LIGHT_STAGE;
    else if (ambientBrightness < darknessThreshold)  stage = DARK_STAGE;
    else                                             stage = INTERSTAGE;
    // (2) Schaltzustand bestimmen
    BrightnessMode mode;
    STORING_Load(STO_BrightnessMode[i], &mode);
    bool setSwitch = BOARD_PowerSwitch(powerSwitch);
    switch (mode)
    {   case BRIGHTNESS_MODE_LIGHT_OFF_DARK_ON:
            if      (stage == LIGHT_STAGE) setSwitch = false;
            else if (stage == DARK_STAGE)  setSwitch = true;
            break;
        case BRIGHTNESS_MODE_LIGHT_ON_DARK_OFF:
            if      (stage == LIGHT_STAGE) setSwitch = true;
            else if (stage == DARK_STAGE)  setSwitch = false;
            break;
        default:
            ;
    }
    // (3) Automatische Abschaltung
    UInt08 time;
    STORING_Load(STO_BrightnessAutoOffTime[i], &time);
    if (time > 0)
    {   static Timer autoOffTimer[] = { TIMER_SECONDS(0), TIMER_SECONDS(0) };
        static UInt08 restarts[]    = { 0, 0 };
        if (setSwitch)
        {   if (TIMER_Duration(&autoOffTimer[i]) == 0)
            {   TIMER_Start(&autoOffTimer[i], 600); // Zeiteinheit: 10 Minuten
                restarts[i] = time-1;
            }
            else if (TIMER_HasExpired(&autoOffTimer[i]))
            {   if (restarts[i] == 0)
                {   setSwitch = false;
                }
                else
                {   restarts[i]--;
                    TIMER_Restart(&autoOffTimer[i]);
                }
            }
        }
        else
        {   TIMER_Start(&autoOffTimer[i], 0);
        }
    }
    // (4) Schalter schalten
    BOARD_SetPowerSwitch(powerSwitch, setSwitch);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedOperationMode(void)
{
    for (UInt08 i=0; i<2; i++)
    {   PowerSwitch powerSwitch = (PowerSwitch)i;
        if (IsExtraOn(powerSwitch)) continue;
        PowerSwitchMode mode;
        STORING_Load(STO_PowerSwitchMode[i], &mode);
        switch (mode)
        {   case POWER_SWITCH_MODE_ON:         BOARD_SetPowerSwitch(powerSwitch, true);  break;
            case POWER_SWITCH_MODE_OFF:        BOARD_SetPowerSwitch(powerSwitch, false); break;
            case POWER_SWITCH_MODE_TIMER:      ProceedTimerAutomatic(powerSwitch);       break;
            case POWER_SWITCH_MODE_BRIGHTNESS: ProceedBrightnessAutomatic(powerSwitch);  break;
            default:
                ;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedExtraOn(void)
{
    for (UInt08 i=0; i<2; i++)
    {   PowerSwitch powerSwitch = (PowerSwitch)i;
        if (!IsExtraOn(powerSwitch)) continue;
        BOARD_SetPowerSwitch(powerSwitch, true);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Startup(void)
{
    static const Cy4ServiceItem  servicesTable[] PROGMEM =
    {   //|                 service                  |          handler           |
        //+----------+------+------------------------+----------------------------+
        { { 0x100100 ,   1  , CY4SERVICE_READ_WRITE }, OnPowerSwitchMode          },
        { { 0x100110 ,   1  , CY4SERVICE_READ_ONLY  }, OnPowerSwitchState         },
        { { 0x100116 ,   1  , CY4SERVICE_WRITE_ONLY }, OnExtraOn                  },
        { { 0x10011A ,   1  , CY4SERVICE_READ_WRITE }, OnExtraOnDuration          },
        { { 0x100120 ,  12  , CY4SERVICE_READ_WRITE }, OnTimerModeParameter       },
        { { 0x100130 ,  12  , CY4SERVICE_READ_WRITE }, OnTimerModeParameter       },
        { { 0x100140 ,  12  , CY4SERVICE_READ_WRITE }, OnTimerModeParameter       },
        { { 0x100180 ,   4  , CY4SERVICE_READ_WRITE }, OnBrightnessModeParameter  },

        { { 0x100200 ,   1  , CY4SERVICE_READ_WRITE }, OnPowerSwitchMode          },
        { { 0x100210 ,   1  , CY4SERVICE_READ_ONLY  }, OnPowerSwitchState         },
        { { 0x100216 ,   1  , CY4SERVICE_WRITE_ONLY }, OnExtraOn                  },
        { { 0x10021A ,   1  , CY4SERVICE_READ_WRITE }, OnExtraOnDuration          },
        { { 0x100220 ,  12  , CY4SERVICE_READ_WRITE }, OnTimerModeParameter       },
        { { 0x100230 ,  12  , CY4SERVICE_READ_WRITE }, OnTimerModeParameter       },
        { { 0x100240 ,  12  , CY4SERVICE_READ_WRITE }, OnTimerModeParameter       },
        { { 0x100280 ,   4  , CY4SERVICE_READ_WRITE }, OnBrightnessModeParameter  },

        { { 0x100410 ,   7  , CY4SERVICE_READ_WRITE }, OnDateTime                 },
        { { 0x100510 ,   1  , CY4SERVICE_READ_WRITE }, OnAmbientBrightness        },
        { { 0x100610 ,   2  , CY4SERVICE_READ_WRITE }, OnAcceptBroardcasts        },
        { { 0x100630 ,  10  , CY4SERVICE_READ_WRITE }, OnServerLink               },
        { { 0xE00020 ,   7  , CY4SERVICE_WRITE_ONLY }, OnDateTime                 },
        { { 0xE00060 ,   1  , CY4SERVICE_WRITE_ONLY }, OnAmbientBrightness        },
        { { 0xFFF020 ,   1  , CY4SERVICE_READ_WRITE }, OnInfoLedSignal            },
        { { 0xFFFF00 ,  61  , CY4SERVICE_READ_WRITE }, OnDeviceLabel              },
        { { 0xFFFF40 ,   1  , CY4SERVICE_WRITE_ONLY }, OnFactorySettings          },
        { { 0xFFFFF0 ,   7  , CY4SERVICE_READ_ONLY  }, OnCallTypePlate            },
        { { 0xFFFFF8 ,   1  , CY4SERVICE_WRITE_ONLY }, OnEnterApplication         },
        { { 0xFFFFFA ,   1  , CY4SERVICE_WRITE_ONLY }, OnEnterProgramLoader       },
        { { 0xFFFFFC ,   2  , CY4SERVICE_WRITE_ONLY }, OnChangeCy4Address         }
    };
    CY4NET_AssignServices(servicesTable);
    CY4NET_SetAddress(STORING_LoadCy4SelfAdr());
    TIMER_Start(&serverQueryInterval[0], CFG_INITIAL_TIME_QUERY_DELAY);
    TIMER_Start(&serverQueryInterval[1], CFG_INITIAL_BRIGHTNESS_REQUEST_DELAY);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Proceed(void)
{
    ProccedServerRequests();
    ProceedLedSignaling();
    ProceedOperationMode();
    ProceedExtraOn();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Shutdown(void)
{
}

//////////////////////////////////////////////////////////////////////////////////////////
