﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Boardmodul: Bereitstellung gerätespezifischer Funktionen und Mechanismen            //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2019 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <util/twi.h>
#include "AVRChip.h"
#include "Timer.h"
#include "Board.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Definitionen  ////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define SW1_SSR     D,2
#define SW1_LED     D,3
#define SW2_SSR     D,4
#define SW2_LED     D,5
#define INFO_LED    D,6

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum                        // Arbeitsanweisungen für das Board-Modul:
{   DO_STARTUP,                     // · Starte die Anwendung neu
    DO_ENTER_PGMLDR,                // · Starte den Programmlader
    DO_PROCEED,                     // · Führe die Arbeit fort
    DO_SHUTDOWN                     // · Schalte das Board ab
} DoNext;

typedef struct                      // Bus-Übertragungsstatus:
{   bool   pending : 1;             // · Übertragung im Gange
    bool   echo    : 1;             // · Echo erhalten
    bool   error   : 1;             // · Übertragung Fehlgeschlagen (Kollision)
    UInt08 byteGap : 5;             // · Vergangene Zeit seit dem letzten Byteempfang [ms]
} BusStatus;

typedef struct                      // Element der Bus-Empfangsschlange:
{   bool   timeout : 1;             // · Zeitüberschreitung bei <byteGap> aufgetreten
    UInt08 padding : 7;             // · Füllbits (ohne Bedeutung)
    UInt08 data    : 8;             // · Datenbyte
} BusRxItem;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static volatile UInt16      msecTicks                             = 0;
static volatile UInt16      secTicks                              = 0;
static volatile UInt08      busTxQueue[CFG_BUS_TX_QUEUE_CAPACITY] = { 0 };
static volatile UInt08      busTxQueueHead                        = 0;
static volatile UInt08      busTxQueueLength                      = 0;
static volatile UInt08      busLoopbackByte                       = 0;
static volatile BusStatus   busStatus                             = { false, false, false, 0};
static volatile BusRxItem   busRxQueue[CFG_BUS_RX_QUEUE_CAPACITY] = { { false } };
static volatile UInt08      busRxQueueHead                        = 0;
static volatile UInt08      busRxQueueTail                        = 0;
static          UInt08      busActivityIndicator                  = 0;
static          DateTime    systemDateTime                        = DATETIME_INVALID;
static          DoNext      doNext                                = DO_STARTUP;

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_PreStart(): Diese Funktion wird einmal nach dem Systemstart aufgerufen. Der   //
//  ----------------  Zeitpunkt ihres Aufrufs macht sie zur einer Sonderfunktion, denn  //
//  ihr Aufruf findet in der Start-Sektion "init3", noch lange vor Aufruf von main(),   //
//  statt. Das ist der richtige Zeitpunkt, um z.B. die Ursache des Resets zu ermitteln  //
//  oder statische Variablen auszulesen, deren Werte vor dem Reset platziert wurden     //
//  und den Neustart überlebt haben. Denn zu diesem Zeitpunkt sind die statischen       //
//  Variablen noch nicht initialisiert.                                                 //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((naked, section(".init3"))) void
BOARD_PreStart(void)
{
	MCUSR = 0;
	wdt_disable();		
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  Reset(): Führt einen physikalischen Reset herbei indem der Watchdog ausgelöst wird. //
//  -------  Reset() kehrt nie zum Aufrufer zurück.                                     //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((noreturn)) static void
Reset(void)
{
    DisableInterrupts();    // Interrupts sperren
    wdt_enable(WDTO_15MS);  // Watchdog-Intervall auf kürzeste Zeit setzen
    __asm__ volatile
    (   "ldi r30, lo8 (%0) \n"      // Nach einem Neustart des Systems wird als
        "ldi r31, hi8 (%0) \n"      // erstes das Programmladermodul ausgeführt.
        "ldi r24, %1       \n"      // Der Programmlader nutzt zur Steuerung seines
        "ldi r25, hi8 (%2) \n"      // Ablaufs Variablen, die in der "Noinit"-Sektion
        "0:                \n"      // angesiedelt sind, d.h. sie werden im Rahmen
        "st  Z+,  r24      \n"      // eines Neustartes nicht initialisiert.
        "cpi r30, lo8 (%2) \n"      // Um nun zu vermeiden, dass übrig gebliebene
        "cpc r31, r25      \n"      // Werte des SRAMs unerwünschte Effekte im
        "brlo 0b"                   // Programmlader hervorrufen, wird vor dem
            :                       // Reset das komplette SRAM mit Nullen überschrieben.
            : "i" (RAMSTART),       // Da durch die Nullung Heap wie Stack komplett
              "i" (0),              // zerstört sind, ist diese Aktion ein Punkt-Ohne-
              "i" (RAMEND+1)        // Wiederkehr.
    );
    while (true);                   // Warten, bis Watchdog-Reset eintritt
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
EnterPgmLdr(void)
{
	if (pgm_read_word(FLASHEND - CFG_PGMLDR_SIZE*1024 + 1) == 0x940C)
	{   //GOTO(FLASHEND - CFG_PGMLDR_SIZE*1024 + 1 + 4);
		// Seit dem ATmega328PB ist FLASHEND definiert als "0x7FFFU". Das abschließende 
		// "U" funktioniert im Inline-Assembler von GOTO nicht mehr. Daher ...
		#if FLASHEND == 0x7FFFU
			GOTO(0x7FFF - CFG_PGMLDR_SIZE*1024 + 1 + 4);
		#else
			#error malformed FLASHEND
		#endif
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Verfahrensfunktionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedSystemTime(void)
{
	if (DATETIME_IsValid(&systemDateTime))
	{   static Timer update = TIMER_MILLISECONDS(1000);
		if (TIMER_RestartIfExpired(&update)) DATETIME_ShiftSecond(&systemDateTime, 1);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Interruptbehandler  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
ISR(TIMER2_COMPA_vect)
{
    // Abweichungskompensation              // Quarztaktbedingt findet der IRQ-Aufruf
    {   static UInt08 counter = 0;          // bei einem Zeitgeber-Zählerwert von 115
        if (counter == 4)                   // nicht exakt jede Millisekunde statt,
        {   counter = 0;                    // sondern etwas schneller:
            OCR2A = F_CPU/(32*1000UL);      //
        }                                   //      115 ÷ 115.2 = 0.9983 ms
        else                                //
        {   counter++;                      // Zwecks Ausgleichs dieser Ungenauigkeit
            OCR2A = F_CPU/(32*1000UL)-1;    // wird bei jedem fünften Aufruf, der
        }                                   // Überlauf um einen Zählerwert verlängert:
    }                                       // 4·(115÷115.2) + (116÷115.2) = 5 ms
    // Ticker inkrementieren
    {   msecTicks++;
	    static UInt16 counter = 0;
	    counter++;
	    if (counter == 1000)
	    {   counter = 0;
		    secTicks++;
	    }
    }
    // Bus Byte-Zeitlücke
    if (busStatus.byteGap <= CFG_BUS_INTERBYTE_TIMEOUT)
    {   busStatus.byteGap++;
	    if (busStatus.byteGap > CFG_BUS_INTERBYTE_TIMEOUT)
	    {   busRxQueue[busRxQueueTail].timeout = true;
	    }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART0_TX_vect)
{                                   // IRQ Abfolge sowie die Signale <echo> und <pending>:
    if (!busStatus.echo)            //
    {   busStatus.error = true;     // Start Tx:    |1|
        busTxQueueLength = 0;       //               :
    }                               // Tx-IRQ:  ···--:-----|2|----|3|-···-|n|----| |-···
    busStatus.echo = false;         // Rx-IRQ:  ···--:-|1|--:-|2|--:--···-- -|n|--:--···
    UInt08 head = busTxQueueHead;   //               :  :___:  :___:          :___:
    if (busTxQueueLength == 0)      // echo:    ...__:__|   |__|   |____...___|   |__...
    {   busStatus.pending = false;  //               :____________________________:
    }                               // pending: ...__|                            |__...
    else
    {   busLoopbackByte = busTxQueue[head];
        UDR0            = busTxQueue[head];
        head++;
        if (head == CARDINALITY(busTxQueue)) head = 0;
        busTxQueueLength--;
    }
    busTxQueueHead = head;
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART0_RX_vect)
{
    UInt08 byte = UDR0;
    busStatus.byteGap = 0;   // Byte empfangen, Zeitzähler zurücksetzen
    if (busStatus.pending && byte == busLoopbackByte) busStatus.echo = true; // Echo
    // Byte in Empfangsschlange stellen
    UInt08 tail = busRxQueueTail;
    busRxQueue[tail].data = byte;
    tail++;
    if (tail == CARDINALITY(busRxQueue)) tail = 0;
    if (tail == busRxQueueHead) busRxQueue[busRxQueueTail].timeout = true; // Überlauf
    else                         busRxQueueTail = tail;
}

//////////////////////////////////////////////////////////////////////////////////////////
// Funktionen ////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_Reset(): Initiiert einen Systemneustart. Die Funktion löst nicht unmittelbar  //
//  -------------  einen Reset aus, sondern führt zu einem Ausstieg aus der Haupt-      //
//  schleife main(). Die danach sukzessiv aufgerufenen "Shutdown"-Funktionen aller      //
//  betroffenen Module bewirken den kontrollierten Abschluss aller Aktivitäten, bevor   //
//  die letzte aufgerufene "Shutdown"-Funktion des Board-Moduls dann den physikali-     //
//  schen Reset auslöst.                                                                //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Reset(void)
{
    doNext = DO_STARTUP;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EnterPgmLdr(): Ähnlich wie BOARD_Reset() initiiert BOARD_EnterPgmLdr() einen  //
//  -------------------  Systemneustart. Im Unterschied zu BOARD_Reset(), wird hier     //
//  jedoch nicht die Anwendung neu gestartet, sondern das System bootet in das Pro-     //
//  grammladermodul.                                                                    //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EnterPgmLdr(void)
{
    doNext = DO_ENTER_PGMLDR;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_MSecTicks(): Die Funktion liefert den Stand eines monotonen Millisekunden-    //
//  -----------------  Zeitzählers. Der Zähler wird interruptgetrieben inkrementiert    //
//  und hat eine Breite von 16-Bit. Kommt es nach 65535 ms zum Überlauf, beginnt der    //
//  Zeitzähler wieder bei 0.                                                            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Aktueller Stand des Millisekunden-Zeitzählers.                             //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
BOARD_MSecTicks(void)
{
    UInt16 ticks = 0;

    Atomic { ticks = msecTicks; }
    return ticks;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SecTicks(): Die Funktion liefert den Stand eines monotonen Sekunden-Zeit-     //
//  ----------------  zählers. Der Zähler wird interruptgetrieben inkrementiert und hat //
//  eine Breite von 16-Bit. Kommt es nach 65535 s (knapp 18¼h) zum Überlauf, beginnt    //
//  der Zeitzähler wieder bei 0.                                                        //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Aktueller Stand des Sekunden-Zeitzählers.                                  //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
BOARD_SecTicks(void)
{
    UInt16 ticks = 0;

    Atomic { ticks = secTicks; }
    return ticks;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EEPLoad(): Kopiert <count> Bytes aus dem EEPROM-Speicher ins RAM.             //
//  ---------------                                                                     //
//                                                                                      //
//  Eingabe: · eepAddress:  Adresse im EEPROM-Speicher.                                 //
//           · ramAddress:  Adresse im Ram.                                             //
//           · count:       Anzahl zu kopierender Bytes.                                //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPLoad(const void* eepAddress, void* ramAddress, UInt16 count)
{
	eeprom_read_block(ramAddress, (void*)eepAddress, count);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EEPSave(): Kopiert <count> Bytes aus dem RAM in den EEPROM-Speicher. Da       //
//  ---------------  EEPROM-Speicher lt. Datenblatt nur ca. 100000 Lösch/Schreibzyklen  //
//  pro Zelle verträgt, wird vor für jedes Bytes vor dem Speichervorgang geprüft, ob    //
//  sich der zu schreibende Wert vom Inhalt der EEP-Speicherzelle unterscheidet. Erst   //
//  danach wird nötigenfalls ein Lösch- und Schreibvorgang durchgeführt.                //
//  Hinweis: Das Beschreiben einer EEPROM-Speicherzelle nimmt mit ca. 8 ms verhältnis-  //
//           mäßig viel Zeit in Anspruch.                                               //
//                                                                                      //
//  Eingabe: · ramAddress:  Adresse im Ram.                                             //
//           · eepAddress:  Adresse im EEPROM-Speicher.                                 //
//           · count:       Anzahl zu kopierender Bytes.                                //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPSave(const void* ramAddress, void* eepAddress, UInt16 count)
{
	eeprom_update_block(ramAddress, (void*)eepAddress, count);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_I2cTransmit(): Sendet <count> Datenbytes aus <buffer> an das I²C-Gerät an 	//
//  -------------------  Adresse <address>.								   			    //
//                                                                                      //
//  Eingabe: · address: Adresse des I²C-Gerätes: 0 .. 127.                              //
//           · buffer:  Speicherbereich mit zu schreibenden Bytes.						//
//           · count:   Anzahl Bytes in <buffer>.										//
//  Ausgabe: 'true': Schreibvorgang war erfolgreich.                                    //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_I2cTransmit(SInt08 address, const UInt08* buffer, UInt08 count)
{
	return BOARD_I2cTransceive(address, buffer, count, NULL, 0);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_I2cReceive(): Liest <count> Datenbytes von dem I²C-Gerät an Adresse <address> //
//  ------------------  und legt die empfangenen Bytes in <buffer> ab.                  //
//                                                                                      //
//  Eingabe: · address: Adresse des I²C-Gerätes: 0 .. 127                               //
//           · buffer:  Speicherbereich, der die zu lesenden Bytes aufnimmt.			//
//           · count:   Anzahl zu lesender Bytes.										//
//  Ausgabe: 'true': Lesevorgang war erfolgreich.                                       //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_I2cReceive(SInt08 address, UInt08* buffer, UInt08 count)
{
	return BOARD_I2cTransceive(address, NULL, 0, buffer, count);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_I2cTransceive(): Sendet als erstes <writeCount> Datenbytes aus <writeBuffer>  //
//  ---------------------  an das I²C-Gerät an Adresse <address> und liest danach    	//
//	<readCount> Datenbytes zurück. Die empfangenen Datenbytes werden in <readBuffer>	//
//  abgelegt. Die Schreib-Lese-Sequenz wird auf auf dem I²C-Bus durch einen "Restart-	//
//	Read"-Zustand getrennt.																//
//	Da Schreib- und Lesevorgänge nacheinander durchgeführt werden, können <writeBuffer>	//
//	und <readBuffer> den gleichen Speicherbereich referenzieren.						//
//                                                                                      //
//  Eingabe: · address: Adresse des I²C-Gerätes: 0 .. 127.                              //
//           · writeBuffer: Speicherbereich mit zu schreibenden Bytes.					//
//           · writeCount:  Anzahl Bytes in <writeBuffer>.                              //
//           · readBuffer:  Speicherbereich, der die zu lesenden Bytes aufnimmt.		//
//           · readCount:   Anzahl zu lesender Bytes.									//
//  Ausgabe: 'true': Schreib-Lese-Vorgang war erfolgreich.                              //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_I2cTransceive(SInt08        address,
                    const UInt08* writeBuffer,
                    UInt08        writeCount,
                    UInt08*       readBuffer,
                    UInt08        readCount)
{
	bool success = true;
	// Schreiben
	if (writeCount > 0)
	{   success = false;
		// I²C Einheit einschalten und Start-Kennung senden
		TWCR0 = BV(TWINT) | BV(TWSTA) | BV(TWEN);
		UInt08 t;
		for (t=100; !(TWCR0 & BV(TWINT)) && t>0; t--) Freeze10us(); // max. 1ms warten
		if (t > 0)
		{   if ((TWSR0 & TW_STATUS_MASK) == TW_START)
			{   // Zielgeräte-Addresse mit Schreib-Anforderung senden (SLA+W)
				TWDR0 = ((UInt08)address << 1) | TW_WRITE;
				TWCR0 = BV(TWINT)|BV(TWEN);
				for (t=100; !(TWCR0 & BV(TWINT)) && t>0; t--) Freeze10us();
				if (t > 0)
				{   if ((TWSR0 & TW_STATUS_MASK) == TW_MT_SLA_ACK)
					{   // <writeCount> Datenbytes übertragen
						UInt08 i;
						for (i=0; i<writeCount; i++)
						{   TWDR0 = writeBuffer[i];
							TWCR0 = BV(TWINT)|BV(TWEN);
							for (t=100; !(TWCR0 & BV(TWINT)) && t>0; t--) Freeze10us();
							if (t == 0 || (TWSR0 & TW_STATUS_MASK) != TW_MT_DATA_ACK) break;
						}
						success = i == writeCount;
					}
				}
			}
		}
	}
	// Lesen
	if (readCount > 0 && success)
	{   // Start/Repeated start-Kennung senden
		TWCR0 = BV(TWINT) | BV(TWSTA) | BV(TWEN);
		UInt08 t;
		for (t=100; !(TWCR0 & BV(TWINT)) && t>0; t--) Freeze10us(); // max. 1ms warten
		if (t > 0)
		{   UInt08 status;
			if (writeCount == 0) status = TW_START;
			else                 status = TW_REP_START;
			if ((TWSR0 & TW_STATUS_MASK) == status)
			{   // Zielgeräte-Addresse mit Leseanforderung übertragen (SLA+R)
				TWDR0 = ((UInt08)address << 1) | TW_READ;
				TWCR0 = BV(TWINT) | BV(TWEN);
				for (t=100; !(TWCR0 & BV(TWINT)) && t>0; t--) Freeze10us();
				if (t > 0)
				{    if ((TWSR0 & TW_STATUS_MASK) == TW_MR_SLA_ACK)
					{   // <readCount> Datenbytes entgegennehmen, das letzte unbestätigt (NACK)
						UInt08 i;
						for (i=0; i<readCount; i++)
						{   UInt08  acknowledge;
							if (i < readCount-1)
							{   TWCR0 = BV(TWINT) | BV(TWEN) | BV(TWEA);
								acknowledge = TW_MR_DATA_ACK;
							}
							else
							{   TWCR0 = BV(TWINT) | BV(TWEN);
								acknowledge = TW_MR_DATA_NACK;
							}
							for (t=100; !(TWCR0 & BV(TWINT)) && t>0; t--) Freeze10us();
							if (t == 0 || (TWSR0 & TW_STATUS_MASK) != acknowledge) break;
							readBuffer[i] = TWDR0;
						}
						success = i == readCount;
					}
				}
			}
		}
	}
	// Stop-Kennung senden und I²C Einheit abschalten
	TWCR0 = BV(TWINT)|BV(TWSTO)|BV(TWEN);
	Freeze100us();
	TWCR0 = 0;
	return success;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SetClock(): Stellt sie Systemuhr mit der angegebenen Datumzeit. Ist die       //
//  ----------------  übergebene Datumzeit ungültig, wird die Systemuhr angehalten.     //
//                                                                                      //
//  Eingabe: · dateTime: Datumzeit, mit der die Systemuhr gestellt wird.                //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetClock(const DateTime* dateTime)
{
	systemDateTime = *dateTime;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_Clock(): Liest die Systemuhr aus. In dem Fall, dass die Systemuhr nicht läuft //
//  -------------  ist die zurückgelieferte Datumzeit ungültig.                         //
//                                                                                      //
//  Eingabe: · dateTime: Referenz, die die Zeitinformation aufnimmt.                    //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Clock(DateTime* dateTime)
{		
	*dateTime = systemDateTime;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsIdle(): Prüft, ob der Bus momentan frei ist. Der Bus gilt als frei, wenn //
//  -----------------  eine Kommunikationspause von mehr als CFG_BUS_INTERBYTE_TIMEOUT  //
//  Millisekunden eintritt.                                                             //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn der Bus frei ist.                                             //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsIdle(void)
{
    bool  idle = false;
    Atomic
    {   idle = !busStatus.pending && busStatus.byteGap > CFG_BUS_INTERBYTE_TIMEOUT;
    }
    return idle;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusTransmit(): Sendet das Datenbyte in <byte> auf den Bus.                    //
//  -------------------  Die Sendeeinheit ist durch folgenden Zustandsautomaten         //
//  realisiert, Ausgangszustand ist IDLE:                                               //
//                                                                                      //
//                                                                                      //
//     +----------+ BOARD_BusTransmit()      +--------------+ BOARD_BusTransmit()       //
//     |//////////+------------------------->|//////////////+--------------------+      //
//     |// IDLE //|                COMPLETED |// TRANSMIT //|                    |      //
//     |//////////|<-------------------------+//////////////|<-------------------+      //
//     +----------+                          +------+-------+                           //
//          /|\                                     | COLLISION                         //
//           |                                      |                                   //
//           |                                     \|/                                  //
//           |                               +-------------+                            //
//           | BOARD_BusTakeFailure()        |/////////////|                            //
//           +-------------------------------+// FAILURE //|                            //
//                                           |/////////////|                            //
//                                           +-------------+                            //
//                                                                                      //
//  Mit Aufruf dieser Funktion wechselt der Automat von IDLE nach TRANSMIT und beginnt  //
//  mit mit der Übertragung des angegebenen Datenbytes. Ist bereits eine Übertragung im //
//  Gange (Zustand TRANSMIT), wird bei Aufruf der Funktion das zu übertragende Daten-   //
//  byte für die spätere Übertragung in die Sende-Warteschlange gestellt. Nach Aus-     //
//  sendung aller zu übertragenden Datenbytes (Ereignis COMPLETED) kehrt der Automat    //
//  wieder in den Zustand IDLE zurück.                                                  //
//  Das Halbduplexverfahren des seriellen Busses bedingt, dass jedes ausgesendete Byte  //
//  unmittelbar als Echo wieder in den Empfangspuffer zurückwandert. Diese Rücküber-    //
//  tragung wird genutzt, um Kollisionen aufzudecken. Unterscheiden sich ausgesendetes  //
//  Byte und Echo, löst das das Ereignis COLLISION aus und damit den Wechsel in den     //
//  Zustand FAILURE. Die Übertragung wird in dem Fall nicht mehr fortgesetzt und die    //
//  Sende-Warteschlange geleert. Gegebenenfalls noch ausstehende Bytes werden also      //
//  nicht mehr übertragen. Im Zustand FAILURE bleiben Aufrufe von BOARD_BusTransmit()   //
//  ohne Wirkung. Um neue Datenbytes senden zu können, muss der Automat erst durch      //
//  Aufruf von BOARD_BusTakeFailure() wieder in den Zustand IDLE versetzt werden.       //
//                                                                                      //
//  Eingabe: · byte: Das zu übertragende Datenbyte.                                     //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_BusTransmit(UInt08 byte)
{
    // Ist die Übertragungsschlange voll, warten, bis wieder Platz ist
    bool  queueFull = false;
    do
    {   Atomic
        {   queueFull = busTxQueueLength == CARDINALITY(busTxQueue);
        }
    } while (queueFull);
    // Byte senden
    Atomic
    {   if (busStatus.error) return; // Zustand FAILURE: Nichts tun
        if (busStatus.pending)
        {   // Zustand TRANSMIT: Byte für Übertragung aufreihen
            UInt08 tail = busTxQueueHead + busTxQueueLength;
            if (tail >= CARDINALITY(busTxQueue)) tail -= CARDINALITY(busTxQueue);
            busTxQueue[tail] = byte;
            busTxQueueLength++;
        }
        else
        {   // Zustand IDLE: Byte unmittelbar senden
            busStatus.pending = true;
            busStatus.echo    = false;
            busLoopbackByte   = byte;
            UDR0              = byte;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsTransmitting(): Prüft, ob sich der Automat (siehe BOARD_BusTransmit())   //
//  -------------------------  im Zustand TRANSMIT befindet. Das ist solange der Fall,  //
//  wie die Sendeeinheit mit der Übertragung von Datenbytes beschäftigt ist.            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn der Sendevorgang noch im Gange ist.                           //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsTransmitting(void)
{
    bool  pending = false;
    Atomic
    {   pending = busStatus.pending;
    }
    return pending;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusTakeFailure(): Prüft, ob sich der Automat (siehe BOARD_BusTransmit()) im   //
//  ----------------------  Zustand FAILURE befindet und löst gegebenenfalls einen      //
//  Wechsel in den Zustand IDLE aus. Übertragungsfehler entstehen in der Regel durch    //
//  Kollisionen auf dem Bus.                                                            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn ein Übertragungsfehler vorlag.                                //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusTakeFailure(void)
{
    bool  error = false;
    Atomic
    {   error = busStatus.error;
        busStatus.error = false;
    }
    return error;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusReceive(): Steht bei Aufruf der Funktion ein empfangenes Datenbyte zur     //
//  ------------------  Abholung bereit, wird es aus der Empfangsschlange entfernt und  //
//  in <byte> übergeben. BOARD_BusReceive() kehrt so lange mit 'true' zurück, wie       //
//  Datenbytes zur Abholung bereitstehen.                                               //
//  Man beachte: Bei einer Unterbrechung der Empfangssequenz kehrt die Funktion mit     //
//  'false' zurück. Umgekehrt gilt das jedoch nicht: Der Rückgabewert 'false' bedeutet  //
//  nicht zwingend, dass auch die Empfangssequenz bereits unterbrochen ist! Dazu siehe  //
//  BOARD_BusIsReceiving().                                                             //
//                                                                                      //
//  Eingabe: · byte: Referenz, die das nächste zur Abholung bereitstehende Datenbyte    //
//                   aufnimmt.                                                          //
//  Ausgabe: 'true', wenn in <byte> ein Datenbyte abgelegt wurde.                       //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusReceive(UInt08* byte)
{
    bool  hasData = false;
    Atomic
    {   UInt08 head = busRxQueueHead;
        if (!busRxQueue[head].timeout)
        {   // Keine Zeitüberschreitungs-Markierung gesetzt
            if (head != busRxQueueTail)
            {   // Datenbyte vorhanden
                *byte = busRxQueue[head].data;
                head++;
                if (head == CARDINALITY(busRxQueue)) head = 0;
                busRxQueueHead = head;
                hasData = true;
            }
        }
    }
    return hasData;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsReceiving(): Prüft, ob die aktuelle Empfangssequenz noch im Gange ist.   //
//  ----------------------                                                              //
//  Eine Gruppe empfangener Datenbytes gilt als zusammenhängende Empfangssequenz, wenn  //
//  die Zeitspanne zwischen dem Empfang sukzessiver Datenbytes einen bestimmten Zeit-   //
//  rahmen nicht überschreitet. Mit anderen Worten: Liegen die Zeitpunkte zweier auf-   //
//  einanderfolgend empfangener Datenbytes länger als CFG_BUS_INTERBYTE_TIMEOUT Milli-  //
//  sekunden auseinander (timeout), gilt die Empfangssequenz als unterbrochen.          //
//                                                                                      //
//    Empfang: ···-|A|-|B|-|C|---------------×---|D|----|E|---------------×--|F|-···    //
//                           :               :            :               :             //
//                           |--- timeout ---|            |--- timeout ---|             //
//                                                                                      //
//    Empfangssequenzen: ABC, DE, F...                                                  //
//                                                                                      //
//  Die Unterbrechungen werden mit in der Empfangsschlange vermerkt, so dass beim Aus-  //
//  lesen der empfangenen Bytes mittels BOARD_BusReceive() die Empfangssequenzen zeit-  //
//  unabhängig separiert werden. Solange BOARD_BusIsReceiving() mit <true> zurückkehrt, //
//  ist die Empfangssequenz noch im Gange.                                              //
//                                                                                      //
//      Beispiel: Empfang von Byte-Sequenzen:                                           //
//                                                                                      //
//       UInt08 sequence[SEQUENCE_CAPACITY];   // SEQUENCE_CAPACITY: 1 .. 255           //
//                                                                                      //
//       UInt08                                                                         //
//       TakeNextSequence(void)                                                         //
//       {                                                                              //
//           static UInt08 index = 0;                                                   //
//                                                                                      //
//           UInt08 length = 0;                                                         //
//           if (BOARD_BusIsReceiving())                                                //
//           {   // Empfang einer Sequenz im Gange                                      //
//               UInt08 byte;                                                           //
//               while (BOARD_BusReceive(&byte)) // Datenbytes abholen und aufsammeln   //
//               {   if (index < sizeof(sequence))                                      //
//                   {   sequence[index] = byte;                                        //
//                       index++;                                                       //
//                   }                                                                  //
//                   else                                                               //
//                   {   // Überlauf ...                                                //
//                   }                                                                  //
//               }                                                                      //
//           }                                                                          //
//           else                                                                       //
//           {    // Empfang einer Sequenz nicht (mehr) im Gange                        //
//                length = index;                                                       //
//               index = 0;                                                             //
//           }                                                                          //
//           return length;   // Rückgabewert: Aktuelle Sequenzlänge                    //
//       }                                                                              //
//                                                                                      //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', solange die Empfangssequenz im Gange ist.                          //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsReceiving(void)
{
    bool  timeout = false;
    Atomic
    {   UInt08 head = busRxQueueHead;
        timeout = busRxQueue[head].timeout;
        if (head != busRxQueueTail) busRxQueue[head].timeout = false;
    }
    return !timeout;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusOnActivity(): Rückruffunktion der Bus-Übertragungseinheit. Die Funktion    //
//  ---------------------  wird aufgerufen, wenn mit dem Gerät über den Bus             //
//  kommuniziert wird.                                                                  //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_BusOnActivity(void)
{
	busActivityIndicator = CFG_BUS_ACT_INDICATOR_DURATION;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusActivity(): Der Bus-Aktivitätsindikator zeigt an, ob aktuell mit dem       //
//  -------------------  Gerät kommuniziert wird. Diese Funktion liefert den Zustand    //
//  des Indikators. Nach Abschluss der Kommunikation bleibt der Indikator noch CFG_-    //
//  BUS_ACT_INDICATOR_LENGTH Millisekunden aktiviert.                                   //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusActivity(void)
{
	return busActivityIndicator != 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetLed(Led led, bool on)
{
	switch (led)
	{	case LED_SWITCH_1: ChgPin(SW1_LED,  on); break;
		case LED_SWITCH_2: ChgPin(SW2_LED,  on); break;
		case LED_INFO:     ChgPin(INFO_LED, on); break;
		default:           ;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_Led(Led led)
{
	switch (led)
	{	case LED_SWITCH_1: return (bool)GetPin(SW1_LED);
		case LED_SWITCH_2: return (bool)GetPin(SW2_LED);
		case LED_INFO:     return (bool)GetPin(INFO_LED);
		default:           return false;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetPowerSwitch(PowerSwitch powerSwitch, bool on)
{
	switch (powerSwitch)
	{	case POWER_SWITCH_1: ChgPin(SW1_SSR, on); break;
		case POWER_SWITCH_2: ChgPin(SW2_SSR, on); break;
		default:             ;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_PowerSwitch(PowerSwitch powerSwitch)
{
	switch (powerSwitch)
	{	case POWER_SWITCH_1: return (bool)GetPin(SW1_SSR);
		case POWER_SWITCH_2: return (bool)GetPin(SW2_SSR);
		default:             return false;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Startup(void)
{
    if (doNext != DO_STARTUP) Reset();
	DisableInterrupts();
    // Watchdog --------------------------------------------------------------------------    
    wdt_enable(WDTO_2S);      
	// Ports'n'Pins ----------------------------------------------------------------------
    ConfPort(B, _L_,    // PB.0: ---     : ---
                _L_,    // PB.1: ---     : ---
                _L_,    // PB.2: ---     : ---
                _L_,    // PB.3: MOSI    : ISP
                _L_,    // PB.4: MISO    : ISP
                _L_,    // PB.5: SCK     : ISP
                _L_,    // PB.6: [XTAL1] : Quarz
                _L_);   // PB.7: [XTAL2] : Quarz
    ConfPort(C, _L_,    // PC.0: ---     : ---
                _L_,    // PC.1: ---     : ---
                _L_,    // PC.2: ---     : ---
                _L_,    // PC.3: ---     : ---
                _L_,    // PC.4: ---     : ---
                _L_,    // PC.5: ---     : ---
                _T_,    // PC.6: RESET	 
                _T_);   //				 
    ConfPort(D, _T_,    // PD.0: BUS_RXD : Bus
                _T_,    // PD.1: BUS_TXD : Bus
                _L_,    // PD.2: SW1_SSR : Solid-State-Relais 1, L = Aus
                _L_,    // PD.3: SW1_LED : Status-LED Schalter 1, L = Aus
                _L_,    // PD.4: SW2_SSR : Solid-State-Relais 2, L = Aus
                _L_,    // PD.5: SW2_LED : Status-LED Schalter 2, L = Aus
                _L_,    // PD.6: INFO_LED: Info-LED, L = Aus
                _L_);   // PD.7: ---     : ---
	ConfPort(E, _L_,    // PE.0: ---     : ---
               	_L_,    // PE.1: ---     : ---
               	_L_,    // PE.2: ---     : ---
              	_L_,    // PE.3: ---     : ---
				_L_,    //
				_L_,    //
				_L_,    //
				_L_);   //				
    // Pause, bis sich alles stabilisiert hat --------------------------------------------
    Freeze100us();
    // Timer 2: Clear Timer on Compare Match (CTC) Mode, Intervall 0.9983 ms -------------
    TCCR2A = BV(WGM21);
    TCCR2B = BV(CS21) | BV(CS20);
    OCR2A  = F_CPU/(32*1000UL)-1;   // 32 = Vorteilung, 1000 = Wiederholrate bei 3.686400
	TIMSK2 = BV(OCIE2A);
    // USART 0: 19200 Baud, 8 Data, 1 Stop, No Parity, Async. ----------------------------
    UBRR0  = F_CPU/16L/19200-1;
    UCSR0A = 0;
    UCSR0B = BV(RXCIE0) | BV(TXCIE0) | BV(RXEN0) | BV(TXEN0);
    UCSR0C = BV(UCSZ01) | BV(UCSZ00);
    // Sonstiges -------------------------------------------------------------------------
    doNext = DO_PROCEED;
    EnableInterrupts();	
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_DoMainLoop(void)
{
    return doNext == DO_PROCEED;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Proceed(void)
{
	ProceedSystemTime();     
	// Bus-Aktivitätsindikator
	UInt16 ticks = BOARD_MSecTicks();
	{	static UInt08 ms = 0;
		if (ms != (UInt08)ticks)
		{	ms = (UInt08)ticks;	
			if (busActivityIndicator > 0)
			{   busActivityIndicator--;
			}
		}
	}
	wdt_reset();	
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Shutdown(void)
{
    while (BOARD_BusIsTransmitting());
    switch (doNext)
    {   case DO_ENTER_PGMLDR: EnterPgmLdr();        break;
        case DO_STARTUP:      Reset();              break;
        default:              ;
    }
    // Abschaltanweisungen ...
}

//////////////////////////////////////////////////////////////////////////////////////////
