﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Datenaufbewahrung                                                                   //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2010 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include "Board.h"
#include "Storing.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Definitionen  ////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define CRC_WEAR_LEVELING_ROTATION  6      // Wechsel zwecks Verschleißreduktion: 1 .. 250
#define NO_CRC_MARKER               0xFFFF // Wertausprägung für "kein CRC"
#define STOP_INDEX                  0xFFFF // Index Wächterwert

//////////////////////////////////////////////////////////////////////////////////////////
//  Konstanten  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
const Storing factoryDef PROGMEM =
{   .STO_DeviceLabel           = "[Power Switch Pair]",
    .STO_InfoLedSignal         = INFOLEDSIGNAL_HEARTBEAT,
    .STO_AcceptBroadcast       = { true, true },
    .STO_ServerLink            = { CY4LINK, CY4LINK },
    .STO_PowerSwitchMode       = { POWER_SWITCH_MODE_OFF, POWER_SWITCH_MODE_OFF },	
	.STO_ExtraOnDuration       = { 5, 5 },
    .STO_AutoTimer             = { { AUTO_TIMER_INVALID, AUTO_TIMER_INVALID, AUTO_TIMER_INVALID },
                                   { AUTO_TIMER_INVALID, AUTO_TIMER_INVALID, AUTO_TIMER_INVALID }
                                 },
    .STO_BrightnessMode        = { BRIGHTNESS_MODE_LIGHT_OFF_DARK_ON, BRIGHTNESS_MODE_LIGHT_OFF_DARK_ON },
    .STO_LightnessThreshold    = { 20, 20 },
    .STO_DarknessThreshold     = { 15, 15 },
    .STO_BrightnessAutoOffTime = {  0,  0 }
};

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static  UInt16  saveIndex = STOP_INDEX;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
UpdateCRC(UInt08 data, UInt16* crc) // 16-Bit CRC nach CCITT
{
    *crc ^= data;
    for (UInt08 i=8; i>0; i--) *crc = (*crc >> 1) ^ ((*crc&1) ? 0x8408 : 0);
}

//////////////////////////////////////////////////////////////////////////////////////////
static UInt16
CalcCRC(void)
{
    UInt16 crc = 0xFFFF;
    for (UInt16 i=0; i<sizeof(Storing); i++)
    {   UInt08  byte;
        BOARD_EEPLoad((const void*)(CFG_EEP_STORING_PARTITION + i), &byte, 1);
        UpdateCRC(byte, &crc);
    }
    if (crc == NO_CRC_MARKER) crc = (UInt16)(~NO_CRC_MARKER); // Prüfsummenwert tauschen
    return crc;
}

//////////////////////////////////////////////////////////////////////////////////////////
static UInt16
LoadCRC(void)
{
    UInt16       crc;
    if (CRC_WEAR_LEVELING_ROTATION < 2)
    {   const void* eepAdr = (const void*)(CFG_EEP_STORING_PARTITION + sizeof(Storing));
        BOARD_EEPLoad(eepAdr, &crc, sizeof(UInt16));
    }
    else
    {   UInt08 i=0;
        do
        {   const void* eepAdr = (const void*)(CFG_EEP_STORING_PARTITION +
                                sizeof(Storing) + i*sizeof(UInt16));
            BOARD_EEPLoad(eepAdr, &crc, sizeof(UInt16));
            i++;
        } while (i <= CRC_WEAR_LEVELING_ROTATION && crc == NO_CRC_MARKER);
    }
    return crc;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void                                            //
SaveCRC(UInt16 crc)                                    //             EEPMEM:
{                                                      //             +---------------+
                                                       //             |               |
    if (crc == NO_CRC_MARKER)                          //             |    Storing    |
    {   crc = (UInt16)(~NO_CRC_MARKER);                //             |     data      |
    }                                                  //             |               |
    void* eepAdr = (void*)(CFG_EEP_STORING_PARTITION + //             +---------------+
                       sizeof(Storing));               //           / | NO_CRC_MARKER | 0
    if (CRC_WEAR_LEVELING_ROTATION < 2)                // CRC      |  |       :       | 1
    {   BOARD_EEPSave(&crc, eepAdr, sizeof(crc));      // Wear-    |  |      CRC      | :
    }                                                  // leveling |  |       :       | :
    else                                               //           \ | NO_CRC_MARKER | n
    {   // Letzte Prüfsumme suchen                     //             +---------------+
        UInt08 i=0;                                    //
        UInt16 tmpCrc;                                 //   n = CRC_WEAR_LEVELING_ROTATION
        while (i <= CRC_WEAR_LEVELING_ROTATION)
        {   BOARD_EEPLoad(eepAdr, &tmpCrc, sizeof(tmpCrc));
            if (tmpCrc != NO_CRC_MARKER) break;
            eepAdr = (void*)((UInt16)eepAdr + sizeof(UInt16));
            i++;
        }
        if (crc == tmpCrc) return;
        // Letzte Prüfsumme löschen (NO_CRC_MARKER)
        tmpCrc = NO_CRC_MARKER;
        BOARD_EEPSave(&tmpCrc, eepAdr, sizeof(tmpCrc));
        // Neue Prüfsumme in nachfolgender Zelle ablegen
        i++;
        if (i > CRC_WEAR_LEVELING_ROTATION) i = 0;
        eepAdr = (void*)(CFG_EEP_STORING_PARTITION + sizeof(Storing) + i*sizeof(crc));
        BOARD_EEPSave(&crc, eepAdr, sizeof(crc));
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Intern  //////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_LoadEEPMem(const void* eepAddress, void* ramAddress, UInt16 count)
{
    BOARD_EEPLoad(eepAddress, ramAddress, count);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_SaveEEPMem(const void* ramAddress, void* eepAddress, UInt16 count)
{
    BOARD_EEPSave(ramAddress, eepAddress, count);
    saveIndex = 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_RestoreDefaults(void)
{
    const void* pgmAddress = &factoryDef;
    void*       eepAddress = (void*)CFG_EEP_STORING_PARTITION;
    for (UInt16 i=0; i<sizeof(Storing); i++)
    {   UInt08 byte = pgm_read_byte(pgmAddress);
        BOARD_EEPSave(&byte, eepAddress, 1);
        pgmAddress = (const void*)((UInt16)pgmAddress + 1);
        eepAddress = (void*)((UInt16)eepAddress + 1);
    }
    saveIndex = 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
Cy4Address
STORING_LoadCy4SelfAdr(void)
{
    UInt08  data[4];
    Cy4Address  address = CY4ADDRESS(CFG_DEFAULT_CY4_ADDRESS_NET,
                                     CFG_DEFAULT_CY4_ADDRESS_DEV);

    BOARD_EEPLoad((const void*)CFG_EEP_CY4_ADDRESS_PARTITION, data, sizeof(data));
    UInt08 invData[2] = { ~data[2], ~data[3] };
    if (data[0] == invData[0] && data[1] == invData[1])
    {   address = CY4ADDRESS_Create(data[0], data[1]);
    }
    return address;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_SaveCy4SelfAdr(const Cy4Address* address)
{
    UInt08  data[4];

    data[0] = CY4ADDRESS_Net(address);
    data[1] = CY4ADDRESS_Dev(address);
    data[2] = ~data[0];
    data[3] = ~data[1];
    BOARD_EEPSave(data, (void*)CFG_EEP_CY4_ADDRESS_PARTITION, sizeof(data));
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_Startup(void)
{
    if (CalcCRC() != LoadCRC()) STORING_RestoreDefaults();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_Proceed(void)
{
    static UInt16  crc = 0;
    UInt08         byte;

    if (saveIndex == STOP_INDEX) return;
    BOARD_EEPLoad((const void*)(CFG_EEP_STORING_PARTITION + saveIndex), &byte, 1);
    if (saveIndex == 0) crc = 0xFFFF;
    UpdateCRC(byte, &crc);
    saveIndex++;
    if (saveIndex == sizeof(Storing))
    {   saveIndex = STOP_INDEX;
        SaveCRC(crc);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_Shutdown(void)
{
    while (saveIndex != STOP_INDEX) STORING_Proceed();
}

//////////////////////////////////////////////////////////////////////////////////////////
